from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3 as NavigationToolbar2
from matplotlib_helpers.points import PointsHandler
from pygtk3_helpers.file_chooser import FileChooserView
from pygtk3_helpers.delegates import SlaveView
from gi.repository import Gtk
from opencv_helpers import imshow, resize
import pandas as pd
import numpy as np
import cv


def load_im(input_file):
    import cv2
    from path_helpers import path

    input_file = path(input_file)
    if input_file.ext.lower() in ('.mp4', '.avi'):
        # Input file is a video, so grab a frame.
        video = cv2.VideoCapture(input_file)
        video.grab()
        success, mat = video.retrieve()
        if not success:
            raise IOError('Error grabbing frame from video.')
        im = cv2.cv.fromarray(mat)
        del video
    elif input_file.ext.lower() in ('.png', '.jpg'):
        # Input file is an image.
        im = cv2.cv.LoadImage(input_file)
    return im


class PointsIOView(SlaveView):
    def create_ui(self):
        self.widget.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.reset_top = Gtk.Button('Reset top')
        self.reset_bottom = Gtk.Button('Reset bottom')
        self.reset_all = Gtk.Button('Reset points')

        for button in (self.reset_top, self.reset_bottom, self.reset_all):
            self.widget.pack_start(button, False, False, 0)

        filters = [{'name': 'Points (*.h5)', 'pattern': ['*.h5']}]
        self.load_points_selector = FileChooserView(label=False, filters=filters)
        self.load_points_selector.show()
        label = Gtk.Label('Load points')
        self.load_points_selector.box.pack_start(label, False, False, 0)
        self.load_points_selector.box.reorder_child(label, 0)

        self.save_points_selector = FileChooserView(label=False,
                                                    action=Gtk.FileChooserAction.SAVE,
                                                    title='Save points to...',
                                                    filters=filters)
        self.save_points_selector.show()
        label = Gtk.Label('Save points')
        self.save_points_selector.box.pack_start(label, False, False, 0)
        self.save_points_selector.box.reorder_child(label, 0)

        self.add_slave(self.save_points_selector)
        self.add_slave(self.load_points_selector)

        # Pack load and save sections to end of row.
        for widget in (self.save_points_selector.widget,
                       self.load_points_selector.widget):
            self.widget.set_child_packing(widget, False, False, 0, Gtk.PackType.END)


class VideoSelectorView(SlaveView):
    def create_ui(self):
        self.widget.set_size_request(640, 640)

        filters = [{'name': 'Video (*.mp4, *.avi)', 'pattern': ['*.mp4', '*.avi']}]
        self.video_selector = FileChooserView(editable=False, filters=filters)
        self.video_selector.show()
        label = Gtk.Label('Select video:')
        self.video_selector.box.pack_start(label, False, False, 0)
        self.video_selector.box.reorder_child(label, 0)
        self.video_selector.on_selected = self.on_selected
        self.add_slave(self.video_selector)
        self.widget.set_child_packing(self.video_selector.widget, False, False, 0,
                                      Gtk.PackType.START)

        self.io_view = PointsIOView()
        self.io_view.show()
        self.add_slave(self.io_view)
        self.io_view.load_points_selector.on_selected = self.on_points_load
        self.io_view.save_points_selector.on_selected = self.on_points_save
        self.io_view.reset_top.connect('clicked', lambda *args: self.reset(0))
        self.io_view.reset_bottom.connect('clicked', lambda *args: self.reset(1))
        self.io_view.reset_all.connect('clicked', lambda *args: self.reset())
        self.widget.set_child_packing(self.io_view.widget, False, False, 0,
                                      Gtk.PackType.START)

        self.registration_view = RegistrationView()
        self.registration_view.show()
        self.add_slave(self.registration_view)

        self.widget.show_all()

    def reset(self, index=None):
        '''
        Reset the points for the specified index position.  If no index is
        specified, reset points for all point handlers.
        '''
        points_handler_count = len(self.registration_view.points)
        if index is None:
            indexes = range(points_handler_count)
        else:
            indexes = [index]

        indexes = [i for i in indexes if i < points_handler_count]

        for i in indexes:
            self.registration_view.points[i].reset()
        if indexes:
            self.registration_view.update_transform()

    def on_points_load(self, value):
        if self.registration_view.axes:
            points = pd.read_hdf(value, '/points')
            self.registration_view.set_points(points)

    def on_points_save(self, value):
        if self.registration_view.axes:
            points = self.registration_view.get_points()
            points.to_hdf(value, '/points', format='t', data_columns=True)

    def on_selected(self, value):
        try:
            self.registration_view.set_image(value)
        except:
            pass


class RegistrationView(SlaveView):
    def __init__(self, *args, **kwargs):
        self.grid = kwargs.pop('grid', GridSpec(2, 1))
        super(RegistrationView, self).__init__(*args, **kwargs)
        self.fig = None
        self.axes = []
        self.im_in = None
        self.im_out = None
        self.points = []

    def create_ui(self):
        self.fig = Figure(figsize=(8, 8), dpi=100)
        self.axes = []

        sw = Gtk.ScrolledWindow()

        plot_box = Gtk.VBox()
        sw.add_with_viewport(plot_box)

        canvas = FigureCanvas(self.fig)
        canvas.set_size_request(500, 600)
        plot_box.pack_start(canvas, True, True, 0)

        toolbar = NavigationToolbar2(canvas, sw)
        self.widget.pack_start(toolbar, False, False, 0)
        self.widget.add(sw)

    def add_subplot(self, *args):
        axis = self.fig.add_subplot(*args)
        self.axes.append(axis)

    def plot(self, index, args, kwargs):
        self.axes[index].plot(*args, **kwargs)
        self.fig.canvas.draw()

    def clear(self):
        try:
            for widget in self.widget:
                self.widget.remove(widget)
        except (Exception, ), why:
            print why

    def refresh(self):
        self.fig.canvas.draw()

    def set_image(self, im, out_shape=None):
        import cv2

        if isinstance(im, str):
            im = load_im(im)
        elif isinstance(im, np.ndarray):
            im = cv2.cv.fromarray(im)
        self.fig.clf()
        self.axes = [self.fig.add_subplot(g) for g in self.grid]
        self.im_in = resize(im, im.width, im.height)
        if out_shape is None:
            out_shape = im.width, im.height
        self.im_out = resize(im, *out_shape)
        imshow(self.im_in, axis=self.axes[0], show_axis=True)
        imshow(self.im_out, axis=self.axes[1], show_axis=True)
        # Use `fig.set_tight_layout(True)` rather than `fig.tight_layout()` to
        # avoid [bug with GTK3 backend][1].
        #
        # [1]: https://github.com/matplotlib/matplotlib/issues/1852
        self.fig.set_tight_layout(True)
        self.points = [PointsHandler(ax) for ax in self.axes]

        # Update transformed image whenever the bounding box in either
        # the input or output frame has been changed.
        def do_update(*args):
            self.update_transform()

        for p in self.points:
            p.connect('box_release_event', do_update)
        self.update_transform()

    def update_transform(self):
        map_mat = cv.CreateMat(3, 3, cv.CV_32FC1)
        cv.GetPerspectiveTransform(map(tuple, self.points[0].points.values),
                                   map(tuple, self.points[1].points.values), map_mat)
        flags = cv.CV_WARP_FILL_OUTLIERS
        cv.WarpPerspective(self.im_in, self.im_out, map_mat, flags=flags)
        imshow(self.im_out, axis=self.axes[1], show_axis=True)
        self.refresh()

    def get_points(self):
        frames = []

        for i, p in enumerate(self.points):
            points_i = p.points.copy()
            points_i.insert(0, 'image_i', i)
            frames.append(points_i)
        return pd.concat(frames)

    def set_points(self, points):
        for i, df_i in points.groupby('image_i'):
            self.points[i].points = df_i[['x', 'y']]
        self.update_transform()

    def reset(self):
        for p in self.points:
            p.reset()



import time
from threading import Thread

import gi
gi.require_version('Gst', '1.0')
from gi.repository import Gst, GstVideo
from path_helpers import path
from .caps import (get_video_source, get_caps_str, get_video_device_key,
                   get_bitrate)


class DrawPipeline(object):
    '''
    Draw video source to window with the specified `xid`.
    '''
    def run(self, xid, device_config=None):
        self.xid = xid
        # Create GStreamer pipeline
        self.pipeline = Gst.Pipeline()

        # Create bus to get events from GStreamer pipeline
        self.bus = self.pipeline.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect('message::error', self.on_error)

        # This is needed to make the video output in our DrawingArea:
        self.bus.enable_sync_message_emission()
        self.bus.connect('sync-message::element', self.on_sync_message)

        # Create GStreamer elements
        if device_config is None:
            self.src = Gst.ElementFactory.make('autovideosrc', 'source')
        else:
            self.src = get_video_source()
            device_key = get_video_device_key()
            self.src.set_property(device_key, device_config['device'])
        self.filter_ = Gst.ElementFactory.make('capsfilter', 'filter')
        self.sink = Gst.ElementFactory.make('autovideosink', 'sink')
        self.sink.set_property('sync', False)
        caps = Gst.Caps(get_caps_str(device_config))
        self.filter_.set_property('caps', caps)

        # Add elements to the pipeline
        self.pipeline.add(self.src)
        self.pipeline.add(self.filter_)
        self.pipeline.add(self.sink)

        self.src.link(self.filter_)
        self.filter_.link(self.sink)
        self.pipeline.set_state(Gst.State.PLAYING)

    def on_sync_message(self, bus, msg):
        if msg.get_structure().get_name() == 'prepare-window-handle':
            print('prepare-window-handle')
            msg.src.set_property('force-aspect-ratio', True)
            msg.src.set_window_handle(self.xid)

    def on_error(self, bus, msg):
        print('on_error():', msg.parse_error())


class RecordPipeline(object):
    def run(self, xid, output_path, device_config=None, bitrate=350 << 3 << 10):
        '''
        Draw video source to window with the specified `xid` and record the
        video to the specified output file path.

        __NB__ The output file container is determined based on the extension
        of the output file path.  Supported containers are `avi` and `mp4`.  In
        either case, the video is encoded in MPEG4 format.

        Arguments
        ---------

         - `xid`: Integer identifier of window to draw frames to.
         - `output_path`: Output file path.
         - `device_config`:
           * Configuration dictionary or a `pandas.Series` in the format of a
             row of a frame returned by `caps.get_device_configs()`.
           * If not provided, the GStreamer `autovideosrc` is used.
         - `bitrate`: Target encode bit rate in bits/second (default=350kB/s)
        '''
        self.xid = xid
        # Create GStreamer pipeline
        self.pipeline = Gst.Pipeline()

        # Create bus to get events from GStreamer pipeline
        self.bus = self.pipeline.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect('message::error', self.on_error)

        # This is needed to make the video output in our DrawingArea:
        self.bus.enable_sync_message_emission()
        self.bus.connect('sync-message::element', self.on_sync_message)

        # Create GStreamer elements
        if device_config is None:
            self.src = Gst.ElementFactory.make('autovideosrc', 'source')
        else:
            self.src = get_video_source()
            device_key = get_video_device_key()
            self.src.set_property(device_key, device_config['device'])
        self.sink = Gst.ElementFactory.make('autovideosink', 'sink')
        self.sink.set_property('sync', False)

        self.filter_ = Gst.ElementFactory.make('capsfilter', 'filter')
        tee = Gst.ElementFactory.make('tee', None)

        sink_queue = Gst.ElementFactory.make('queue', None)
        capture_queue = Gst.ElementFactory.make('queue', None)
        encoder = Gst.ElementFactory.make('avenc_mpeg4', None)
        encoder.set_property('bitrate', bitrate)
        encoder.set_property('bitrate-tolerance', 500 << 10)
        if path(output_path).ext.lower() == '.mp4':
            muxer = Gst.ElementFactory.make('mp4mux', None)
        elif path(output_path).ext.lower() == '.avi':
            muxer = Gst.ElementFactory.make('avimux', None)
        else:
            raise ValueError('Unsupported output file type: %s' %
                             path(output_path).ext)
        filesink = Gst.ElementFactory.make('filesink', None)
        filesink.set_property('location', output_path)

        videorate = Gst.ElementFactory.make('videorate', None)
        filter1 = Gst.ElementFactory.make('capsfilter', None)
        filter1.set_property('caps',
                             Gst.Caps('video/x-raw,framerate={framerate_numerator}/{framerate_denominator}'
                                      .format(**device_config)))

        caps = Gst.Caps(get_caps_str(device_config))
        self.filter_.set_property('caps', caps)

        src_elements = (self.src, self.filter_, videorate, filter1, tee)
        sink_elements = (sink_queue, self.sink)
        capture_elements = (capture_queue, encoder, muxer, filesink)

        # Add elements to the pipeline
        for d in src_elements + sink_elements + capture_elements:
            self.pipeline.add(d)

        for elements in (src_elements, sink_elements, capture_elements):
            for i, j in zip(elements[:-1], elements[1:]):
                i.link(j)

        tee.link(sink_elements[0])
        tee.link(capture_elements[0])


        self.output_path = output_path
        self.tee = tee
        self.muxer = muxer
        self.src_elements = src_elements
        self.sink_elements = sink_elements
        self.capture_elements = capture_elements

        self.pipeline.set_state(Gst.State.PLAYING)
        self._alive = True

    def on_sync_message(self, bus, msg):
        if msg.get_structure().get_name() == 'prepare-window-handle':
            print('prepare-window-handle')
            msg.src.set_property('force-aspect-ratio', True)
            msg.src.set_window_handle(self.xid)

    def on_error(self, bus, msg):
        print('on_error():', msg.parse_error())

    def eos_callback(self, pad, event):
        self.muxer.get_static_pad('src').remove_probe(self.eos_probe)
        self._alive = False
        return True

    def block_callback(self, pad, event):
        mux_pad = self.muxer.get_static_pad('src')
        capture_pad = self.capture_elements[0].get_static_pad('sink')
        self.eos_probe = mux_pad.add_probe(Gst.PadProbeType.EVENT_DOWNSTREAM,
                                           self.eos_callback)
        capture_pad.send_event(Gst.Event.new_eos())
        self.tee.pads[2].remove_probe(self.block_probe)
        return True

    def stop(self):
        # Start callback chain to send EOS (end of stream) event to video
        # muxer.  This is required, for example, when recording to `mp4`, where
        # the EOS event triggers the muxer to write the video header to the
        # file.  See [here][1] for more information.
        #
        # The basic idea is to:
        #
        #  - Block the tee source pad for the capture branch
        #  - Send EOS event through the capture queue `sink` pad
        #  - Wait until EOS is received by the muxer `sink` pad
        #  - Stop the pipeline (wait is complete when `self._alive` is `True`)
        #
        # [1]: http://gstreamer.freedesktop.org/data/doc/gstreamer/head/manual/html/section-dynamic-pipelines.html#section-dynamic-changing
        self.block_probe = self.tee.pads[2].add_probe(Gst.PadProbeType.BLOCK_DOWNSTREAM,
                                                      self.block_callback)
        for i in range(10):
            if not self._alive:
                self.pipeline.set_state(Gst.State.NULL)
                break
            time.sleep(.2)


class PipelineManager(object):
    def __init__(self):
        self.pipeline = None
        self.active_config = None

    def set_config(self, xid, device_config, record_path=None):
        print get_caps_str(device_config)

        self._stop()
        self.active_config = device_config  # = configs.iloc[config_index]

        kwargs = {'device_config': device_config}
        if record_path is not None:
            self.pipeline = RecordPipeline()
            kwargs['output_path'] = record_path
            kwargs['bitrate'] = get_bitrate(device_config.height)
        else:
            self.pipeline = DrawPipeline()

        gst_thread = Thread(target=self.pipeline.run, args=(xid, ), kwargs=kwargs)
        gst_thread.daemon = True
        gst_thread.start()
        gst_thread.join()

    def _stop(self):
        if self.pipeline is not None and hasattr(self.pipeline, 'pipeline'):
            if hasattr(self.pipeline, 'stop'):
                self.pipeline.stop()
            else:
                self.pipeline.pipeline.set_state(Gst.State.NULL)
            time.sleep(0.05)

    def __del__(self):
        self._stop()



import re
import platform
import time
from threading import Thread

from gi.repository import GObject, Gtk
if platform.system() == 'Linux':
    from gi.repository import GdkX11
from pygtk3_helpers.delegates import SlaveView
from pygtk3_helpers.file_chooser import FileChooserView
from path_helpers import path
from .pipeline import PipelineManager
from .caps import get_device_configs


GObject.threads_init()


class VideoModeSelector(SlaveView):
    def __init__(self, configs):
        if configs is None:
            self.configs = get_device_configs()
        else:
            self.configs = configs
        super(VideoModeSelector, self).__init__()

    def set_configs(self, configs):
        config_str_f = lambda c: '[{label}] {width}x{height}\t{framerate:.0f}fps'.format(**c)

        self.config_store.clear()
        for i, config_i in configs.iterrows():
            self.config_store.append([i, config_i, config_str_f(config_i)])

    def create_ui(self):
        self.config_store = Gtk.ListStore(int, object, str)
        self.set_configs(self.configs)

        self.config_combo = Gtk.ComboBox.new_with_model(self.config_store)
        renderer_text = Gtk.CellRendererText()
        self.config_combo.pack_start(renderer_text, True)
        self.config_combo.add_attribute(renderer_text, "text", 2)
        self.config_combo.connect("changed", self.on_config_combo_changed)
        self.widget.pack_start(self.config_combo, False, False, 0)

    def on_config_combo_changed(self, combo):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            row_id, config = model[tree_iter][:2]
            self.on_config_selected(row_id, config)

    def on_config_selected(self, row_id, config):
        pass


class VideoArea(SlaveView):
    def __init__(self, width=None, height=None):
        self.xid = None

    def create_ui(self):
        self.drawingarea = Gtk.DrawingArea()
        self.drawingarea.set_size_request(480, 270)
        self.xid = self.drawingarea.get_property('window').get_xid()
        self.widget.pack_start(self.drawingarea, False, False, 0)

    def run(self):
        self.__run()
        self.window.show_all()
        # You need to get the XID after window.show_all().  You shouldn't get it
        # in the on_sync_message() handler because threading issues will cause
        # segfaults there.
        Gtk.main()

    def quit(self, window):
        Gtk.main_quit()


class VideoView(SlaveView):
    """
    SlaveView for displaying GStreamer video sink
    """

    def __init__(self, width=480, height=270):
        self.width = width
        self.height = height
        self.xid = None
        super(VideoView, self).__init__()

    def create_ui(self):
        self.drawingarea = Gtk.DrawingArea()
        self.drawingarea.set_size_request(self.width, self.height)
        self.drawingarea.connect('realize', self.on_realize)
        self.drawingarea.show()
        self.widget.pack_start(self.drawingarea, True, True, 0)

    def on_realize(self, widget):
        self.xid = self.drawingarea.get_property('window').get_xid()


class RecordView(SlaveView):
    def __init__(self, device_configs=None):
        self.video_view = None
        super(RecordView, self).__init__()
        if device_configs is None:
            self.device_configs = get_device_configs()
            self.device_configs = self.device_configs[(self.device_configs
                                                       .format == 'I420')
                                                      & (self.device_configs
                                                         .framerate > 10)]
        else:
            self.device_configs = device_configs
        self.pipeline_manager = PipelineManager()
        self.config_requested = None
        self.record_path = None

    def create_ui(self):
        self.record_control = RecordControl()
        self.video_view = VideoView()
        for slave in (self.record_control, self.video_view):
            slave.show()
            self.add_slave(slave)
        self.record_control.on_changed = self.on_options_changed

        # Pack load and save sections to end of row.
        self.widget.set_child_packing(self.record_control.widget, False, False,
                                      0, Gtk.PackType.START)

        def update_gui(view):
            while True:
                view.refresh_config()
                time.sleep(1.)

        self.pipeline_thread = Thread(target=update_gui, args=(self, ))
        self.pipeline_thread.daemon = True
        self.pipeline_thread.start()

    def on_options_changed(self, config, record_path):
        self.config_requested = config
        self.record_path = record_path

    def refresh_config(self):
        '''
        __NB__ This *must* be called from a *different* thread than the GUI/Gtk thread.
        '''
        if self.config_requested is not None:
            while self.video_view.xid is None:
                print 'waiting for GUI...'
                time.sleep(1)
            self.pipeline_manager.set_config(self.video_view.xid,
                                             self.config_requested,
                                             record_path=self.record_path)
            self.config_requested = None


class RecordControl(SlaveView):
    def __init__(self, device_configs=None):
        super(RecordControl, self).__init__()
        if device_configs is None:
            self.device_configs = get_device_configs()
            self.device_configs = self.device_configs[self.device_configs.format == 'I420']
        else:
            self.device_configs = device_configs
        self.config = None
        self._record_path = None

    def create_ui(self):
        self.mode_selector = VideoModeSelector(self.device_configs)
        self.widget.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.record_button = Gtk.CheckButton('Record')
        self.record_button.connect('toggled', self.on_record_toggled)
        self.record_button.show()

        self.auto_increment_button = Gtk.CheckButton('Auto-increment')
        self.auto_increment_button.show()

        filters = [{'name': 'MP4 (*.mp4)', 'pattern': ['*.mp4']},
                   {'name': 'AVI (*.avi)', 'pattern': ['*.avi']}]
        self.record_path_selector = FileChooserView(editable=True,
                                                    action=Gtk.FileChooserAction.SAVE,
                                                    title='Record to...',
                                                    filters=filters)
        self.record_path_selector.show()

        self.mode_selector.on_config_selected = self.on_config_selected
        self.mode_selector.show()
        self.add_slave(self.mode_selector)
        self.add_slave(self.record_path_selector)
        self.widget.pack_start(self.record_button, False, False, 0)
        self.widget.pack_start(self.auto_increment_button, False, False, 0)

        # Pack load and save sections to end of row.
        self.widget.set_child_packing(self.mode_selector.widget, False, False, 0,
                                      Gtk.PackType.START)

    @property
    def record_path(self):
        '''
        If recording is not enabled, return `None` as record path.
        '''
        if self.record_button.get_property('active') and (self.record_path_selector
                                                          .selected_path):
            return self.record_path_selector.selected_path
        else:
            return None

    @record_path.setter
    def record_path(self, value):
        '''
        If recording is not enabled, return `None` as record path.
        '''
        if value is not None:
            self.record_path_selector.selected_path = value

    def on_record_toggled(self, button):
        auto_increment = self.auto_increment_button.get_property('active')
        if auto_increment and self.record_path:
            record_path = path(self.record_path)
            namebase = record_path.namebase
            numeric_match = re.search(r'(?P<name>.*?)(?P<zero_pad>0*)'
                                      '(?P<count>\d+)$', namebase)
            if numeric_match:
                name = numeric_match.group('name')
                count = int(numeric_match.group('count')) + 1
            else:
                name = namebase
                count = 1
            namebase = '{name}{count}'.format(name=name, count=count)
            record_path = record_path.parent.joinpath(namebase +
                                                      record_path.ext)
            self.record_path = str(record_path)
        self.on_changed(self.config, self.record_path)

    def on_config_selected(self, row_id, config):
        self.config = config
        self.on_changed(self.config, self.record_path)

    def on_changed(self, config, record_path):
        pass



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import sys
sys.path.insert(0, '.')
import version


setup(name='webcam-recorder',
      version=version.getVersion(),
      author='Christian Fobel',
      author_email='christian@fobel.net',
      url='https://github.com/cfobel/webcam-recorder',
      license='LGPL-3.0',
      packages=['webcam_recorder'])



# coding: utf-8
import platform
import re
from threading import Thread

import gi
gi.require_version('Gst', '1.0')
from gi.repository import Gst
import pandas as pd
import numpy as np
from path_helpers import path


class DeviceNotFound(Exception):
    pass


# Use recommended bitrates from YouTube:
#
#     2160p (4k)    35-45 Mbps
#     1440p (2k)    16 Mbps
#     1080p         8 Mbps
#     720p          5 Mbps
#     480p          2.5 Mbps
#     360p          1 Mbps
#
# [1]: https://support.google.com/youtube/answer/1722171?hl=en
BITRATES = (pd.Series([35, 16, 8, 5, 2.5, 1],
                      index=[2160, 1440, 1080, 720, 480, 360],
                      name='bits_per_second') * (1 << 20)).astype(int)
BITRATES.index.name = 'height'


def get_bitrate(height):
    return BITRATES[BITRATES.index >= height].iloc[-1]


def get_video_device_key():
    if platform.system() == 'Linux':
        device_key = 'device'
    elif platform.system() == 'Windows':
        device_key = 'device-name'
    else:
        raise ValueError('Unsupported platform: %s' % platform.system())
    return device_key


def get_video_sources():
    if platform.system() == 'Linux':
        try:
            devices = path('/dev/v4l/by-id').listdir()
        except OSError:
            raise DeviceNotFound, 'No devices available'
    else:
#         try:
#             devices = GstVideoSourceManager.get_video_source().probe_get_values_name(
#                     'device-name')
#         except:
#             devices = []
#         if not devices:
#             raise DeviceNotFound, 'No devices available'
#         device_key = 'device-name'
        raise ValueError('Unsupported platform: %s' % platform.system())
    return devices


def get_caps_str(device_config):
    return ('video/x-raw, format=(string){format}, width=(int){width}, '
            'height=(int){height}, framerate=(fraction){framerate_numerator}/'
            '{framerate_denominator}').format(**device_config)


def caps_str_to_df(caps_str):
    '''
    Parse caps string (as returned by `Gst.Pad.query_caps().to_string()`) into
    `pandas.DataFrame` table, with one row per configuration.
    '''
    structures = [dict([process_dict(v.groupdict())
                   for v in re.finditer(
                       r'(?P<key>[^ ]*)=\((?P<type>.*?)\)((?P<value>[^{].*?)|{(?P<values>.*?)})(,|$)', s)])
                  for s in caps_str.split(';')]

    df = pd.DataFrame(structures)
    df.reset_index(drop=True, inplace=True)

    def compute_multi(df):
        multi_values = [[(c, k) for k, v in df[c].iteritems()
                         if isinstance(v, list)] for c in df]
        value_lists = [m for m in multi_values if m]
        if value_lists:
            return pd.DataFrame(np.concatenate(value_lists),
                             columns=['label', 'index'])
        else:
            return pd.DataFrame()

    df_multi = compute_multi(df)

    while df_multi.shape[0] > 0:
        df = resolve_multi(df, df_multi)
        df_multi = compute_multi(df)

    if 'framerate' in df:
        df['framerate_numerator'] = df['framerate'].map(lambda v: v.num)
        df['framerate_denominator'] = df['framerate'].map(lambda v: v.denom)
        df['framerate'] = df['framerate_numerator'] / df['framerate_denominator']
    if 'pixel-aspect-ratio' in df:
        df['pixel-aspect-ratio_numerator'] = df['pixel-aspect-ratio'].map(lambda v: v.num)
        df['pixel-aspect-ratio_denominator'] = df['pixel-aspect-ratio'].map(lambda v: v.denom)
        df['pixel-aspect-ratio'] = (df['pixel-aspect-ratio_numerator'] /
                                    df['pixel-aspect-ratio_numerator'])

    return df.sort(['framerate', 'width', 'height']).reset_index(drop=True)


def resolve_multi(df, df_multi):
    for label, indexes in df_multi.groupby('label')['index']:
        values = np.concatenate(df[label].map(lambda v: [v]
                                              if not isinstance(v, list) else v))
        value_count = df[label].map(lambda v: len(v) if isinstance(v, list) else 1)
        del df[label]
        index = np.concatenate([np.repeat(v, n) for n, v in zip(value_count, df.index)])
        df2 = pd.DataFrame(values, index=index, columns=[label])
        break
    result = df2.join(df).reset_index(drop=True)
    if 'index' in result:
        result = result.drop('index', axis=1)
    return result


def translate(v, dtype):
    if dtype == 'fraction':
        return Gst.Fraction(*map(int, v.split('/')))
    elif dtype == 'int':
        return int(v)
    elif dtype == 'float':
        return int(v)
    elif dtype == 'string':
        return v.strip()
    else:
        raise TypeError('Unsupported type: %s' % dtype)


def process_dict(d):
    if d['values'] is not None:
        # A list of values was provided.
        value = map(lambda v: translate(v, dtype=d['type']),
                    d['values'].strip().split(','))
    else:
        value = translate(d['value'], dtype=d['type'])
    return d['key'], value


def get_video_source():
    if platform.system() == 'Linux':
        video_source = Gst.ElementFactory.make('v4l2src', 'video_source')
    else:
        video_source = Gst.ElementFactory.make('dshowvideosrc', 'video_source')
    return video_source


class VideoSourceCaps(object):
    def run(self, video_device=None):
        # Create GStreamer pipeline
        self.pipeline = Gst.Pipeline()

        self.errors = []
        # Create bus to get events from GStreamer pipeline
        self.bus = self.pipeline.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect('message::error', self.on_error)

        # Create GStreamer elements
        if video_device is None:
            self.src = Gst.ElementFactory.make('autovideosrc', None)
        else:
            self.src = get_video_source()
            device_key = get_video_device_key()
            self.src.set_property(device_key, video_device)

        # Add elements to the pipeline
        self.pipeline.add(self.src)

        self.pipeline.set_state(Gst.State.READY)

        pad = self.src.get_static_pad('src')
        caps = pad.query_caps()
        scaps = caps.simplify()

        self.df_caps = caps_str_to_df(scaps.to_string())

    def stop(self):
        self.pipeline.set_state(Gst.State.NULL)

    def __del__(self):
        self.stop()
        del self.pipeline

    def on_error(self, bus, msg):
        self.errors.append(msg.parse_error())


def get_device_configs():
    '''
    Return a `pandas.DataFrame`, where each row corresponds to an available
    device configuration, including the `device` (i.e., the name of the
    device).
    '''
    frames = []

    for i in range(2):
        for device in get_video_sources():
            df_device_i = get_configs(device)
            df_device_i.insert(0, 'device', str(device))
            frames.append(df_device_i)

    device_configs = pd.concat(frames).drop_duplicates()
    device_configs['label'] = device_configs.device.map(
        lambda x: x.split('/')[-1].split('-')[1].split('_')[0])
    device_configs['bitrate'] = device_configs.height.map(get_bitrate)
    return device_configs


def get_configs(device):
    caps = VideoSourceCaps()
    caps.run(device)
    df_device_i = caps.df_caps.dropna().copy()
    return df_device_i



import os
import pandas as pd
import numpy as np
import gi
gi.require_version('Gst', '1.0')
from gi.repository import GObject, Gst, Gtk, Gdk, GdkPixbuf, GstVideo
from .video_view import RecordView


if __name__ == '__main__':
    GObject.threads_init()
    Gst.init(None)

    view = RecordView()
    view.prepare_ui()
    view.widget.connect('destroy', lambda *args: view.hide_and_quit())
    view.show_and_run()
